"""
A style's layers property lists all the layers available in that style.
The type of layer is specified by the "type" property, and must be one of background,
fill, line, symbol, raster, circle, fill-extrusion, heatmap, hillshade.

Except for layers of the background type, each layer needs to refer to a source. Layers take the data that they get from a source, optionally filter features, and then define how those features are styled.
"""

from enum import Enum
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field, validator

Expression = List[Any]
id = str
layout = Dict
Color = str


class TypeEnum(str, Enum):
    fill = "fill"
    line = "line"
    symbol = "symbol"
    circle = "circle"
    heatmap = "heatmap"
    fillExtrusion = "fill-extrusion"
    raster = "raster"
    hillshade = "hillshade"
    background = "background"


class VisibleEnum(str, Enum):
    """
    Layout property. Optional enum.
    One of "visible", "none". Defaults to "visible".
    Whether this layer is displayed."""

    visible = "visible"
    none = "none"


class Paint(Dict[str, Any]):
    ...


class Layout(Dict[str, Any]):
    ...


class BackgroundPaint(BaseModel):
    backgroundColor: Optional[Color] = Field(alias="background-color")
    backgroundOpacity: Optional[int] = Field(alias="background-opacity")
    backgroundPattern: Optional[str] = Field(alias="background-pattern")


class Layer(BaseModel):
    """
    Layers have two sub-properties that determine how data from that layer is rendered: layout and paint properties.
    """

    type: Literal["fill", "line", "symbol", "circle", "heatmap", "fill-extrusion", "raster", "hillshade", "background"]
    filter: Optional[Expression] = Field(
        description="""A expression specifying conditions on source features. Only features that match the filter are displayed. Zoom expressions in filters are only evaluated at integer zoom levels. The feature-state expression is not supported in filter expressions."""
    )
    id: str = Field(description="""Unique layer name.""")
    maxzoom: Optional[int] = Field(
        description="""
        Optional number between 0 and 24 inclusive. The maximum zoom level for the layer. At zoom levels equal to or greater than the maxzoom, the layer will be hidden.
    """
    )
    metadata: Optional[Dict[str, Any]] = Field(
        description="""
        Arbitrary properties useful to track with the layer, but do not influence rendering. Properties should be prefixed to avoid collisions, like 'mapbox:'.
    """
    )
    minzoom: Optional[int] = Field(
        description="""
    The minimum zoom level for the layer. At zoom levels less than the minzoom, the layer will be hidden.
    """
    )
    source: Optional[str] = Field(description="""Name of a source description to be used for this layer. Required for all layer types except background.""")
    sourceLayer: Optional[str] = Field(
        alias="source-layer",
        description="""Layer to use from a vector tile source. Required for vector tile sources; prohibited for all other source types, including GeoJSON sources.""",
    )
    layout: Optional[Layout] = Field(description="Layout propertes for the layer")
    paint: Optional[Paint]

    class Config:
        allow_population_by_field_name = True  # This is true to permit sourceLayer to be set

    @validator("source", allow_reuse=True, always=True)
    def is_background_or_source(cls, v, values):
        if values["type"] == "background":
            return
        if v is not None:
            return v
        raise AssertionError("Layers except 'background' type require a source")



from ast import Expression
from typing import List, Union

from django.contrib.gis.db import models
from django.db.models import F, Func, Value


class JSONObject(Func):
    """
    This is backported from Django 4
    to provide an easier way to generate the "Properties" dict
    """

    function = "JSONB_BUILD_OBJECT"
    output_field = models.JSONField()

    def __init__(self, **fields: Union[F, Expression, str]):
        expressions = []  # type: List[Union[Value, Union[F, Expression, str]]]
        for key, value in fields.items():
            expressions.extend((Value(key), value))
        super().__init__(*expressions)

    def as_postgresql(self, compiler, connection, **extra_context):
        return self.as_sql(
            compiler,
            connection,
            function="JSONB_BUILD_OBJECT",
            **extra_context,
        )



from enum import Enum
from typing import Dict, List, Literal, Optional, Tuple, Union

from pydantic import AnyUrl, BaseModel, Field, HttpUrl
from typing_extensions import Annotated

from djangostreetmap.annotations import GeoJsonFeature, GeoJsonFeatureCollection

from . import layer

"""
Sources state which data the map should display.
Specify the type of source with the "type" property,
which must be one of vector, raster, raster-dem, geojson,
image, video. Adding a source isn't enough to make data
appear on the map because sources don't contain styling
details like color or width. Layers refer to a source
and give it a visual representation. This makes it possible
to style the same source in different ways,
like differentiating between types of roads in
a highways layer.
"""

XY = Tuple[float, float]
Coords = Tuple[XY, XY, XY, XY]


class SchemeEnum(str, Enum):
    xyz = "xyz"  # Slippy map tilenames scheme.
    tms = "tms"  # OSGeo spec scheme.


class EncodingEnum(str, Enum):
    terrarium = "terrarium"
    mapbox = "mapbox"


class Source(BaseModel):
    attribution: Optional[str] = Field(description="Contains an attribution to be displayed when the map is shown to a user.")
    bounds: List[float] = Field(
        [-180, -85.051129, 180, 85.051129],
        description="An array containing the longitude and latitude of the southwest and northeast corners of the source's bounding box in the following order: Field(description="
        ") [sw.lng, sw.lat, ne.lng, ne.lat]. When this property is included in a source, no tiles outside of the given bounds are requested by MapLibre GL.",
    )
    maxzoom: int = Field(
        22,
        description="Maximum zoom level for which tiles are available, as in the TileJSON spec. Data from tiles at the maxzoom are used when displaying the map at higher zoom levels.",
    )
    minzoom: int = Field(0, description="Minimum zoom level for which tiles are available, as in the TileJSON spec.")
    tiles: Optional[List[AnyUrl]] = Field(default_factory=list, description="An array of one or more tile source URLs, as in the TileJSON spec.")
    url: Optional[AnyUrl] = Field(None, description="A URL to a TileJSON resource.")
    volatile: bool = Field(False, description="A setting to determine whether a source's tiles are cached locally.")


class Vector(Source):
    """
    A vector tile source. Tiles must be in Mapbox Vector
    Tile format. All geometric coordinates in vector tiles
    must be between -1 * extent and (extent * 2) - 1 inclusive.
    All layers that use a vector source must specify a
    "source-layer" value. For vector tiles hosted by Mapbox,
    the "url" value should be of the form mapbox://tilesetid.
    """

    type: Literal["vector"] = "vector"
    promoteId: Optional[str] = Field(
        description="""
        A property to use as a feature id (for feature state).
        Either a property name, or an object of the
        form {<sourceLayer>: Field(description="") <propertyName>}.
        If specified as a string for a vector tile source,
        the same property is used across all its source layers."""
    )
    url: Optional[AnyUrl] = Field(None, description="A URL to a TileJSON resource.")
    scheme: SchemeEnum = Field(SchemeEnum.xyz, description="Influences the y direction of the tile coordinates. The global-mercator (aka Spherical Mercator) profile is assumed.")


class Raster(Source):
    type: Literal["raster"] = "raster"
    tileSize: int = Field(512, description="The minimum visual size to display tiles for this layer. Only configurable for raster layers.")


class RasterDem(Source):
    """
    A raster DEM source. Only supports Mapbox Terrain RGB (mapbox://mapbox.terrain-rgb):
    """

    type: Literal["raster-dem"] = "raster-dem"
    encoding: EncodingEnum = Field(EncodingEnum.mapbox, description="The encoding used by this source. Mapbox Terrain RGB is used by default")
    url: Optional[AnyUrl] = Field(None, description="A URL to a TileJSON resource.")


class GeoJson(Source):
    """
    A GeoJSON source. Data must be provided via a "data" property, whose value can be a URL or inline GeoJSON.
    """

    type: Literal["geojson"] = "geojson"
    buffer: int = Field(
        128,
        description="Size of the tile buffer on each side. A value of 0 produces no buffer. A value of 512 produces a buffer as wide as the tile itself. Larger values produce fewer rendering artifacts near tile edges and slower performance.",
    )
    cluster: bool = Field(
        False,
        description="""
    If the data is a collection of point features, setting this to true clusters the points by radius into groups.
    Cluster groups become new Point features in the source with additional properties:
        cluster Is true if the point is a cluster
        cluster_id A unqiue id for the cluster to be used in conjunction with the cluster inspection methods
        point_count Number of original points grouped into this cluster
        point_count_abbreviated An abbreviated point count
    """,
    )
    clusterMaxZoom: Optional[int] = Field(
        description="Max zoom on which to cluster points if clustering is enabled. Defaults to one zoom less than maxzoom (so that last zoom features are not clustered). Clusters are re-evaluated at integer zoom levels so setting clusterMaxZoom to 14 means the clusters will be displayed until z15."
    )
    clusterMinPoints: Optional[int] = Field(2, description="Minimum number of points necessary to form a cluster if clustering is enabled.")
    clusterProperties: Optional[Dict[str, layer.Expression]] = Field(
        description="""
        An object defining custom properties on the generated clusters
        if clustering is enabled, aggregating values from clustered points.
        Has the form {"property_name": [operator, map_expression]}. operator
        is any expression function that accepts at least 2 operands (e.g. "+" or "max") —
        it accumulates the property value from clusters/points the cluster contains;
        map_expression produces the value of a single point.
        Example: {"sum": ["+", ["get", "scalerank"]]}. For more advanced use cases,
        in place of operator, you can use a custom reduce expression that references a
        special ["accumulated"] value,
        e.g.: {"sum": [["+", ["accumulated"], ["get", "sum"]], ["get", "scalerank"]]}
    """
    )
    clusterRadius: int = Field(50, description="Radius of each cluster if clustering is enabled. A value of 512 indicates a radius equal to the width of a tile.")
    data: Union[str, GeoJsonFeature, GeoJsonFeatureCollection] = Field(description="A URL to a GeoJSON file, or inline GeoJSON.")
    filter: Optional[str] = Field(description="An expression for filtering features prior to processing them for rendering.")
    generateId: bool = Field(
        False,
        description="Whether to generate ids for the geojson features. When enabled, the feature.id property will be auto assigned based on its index in the features array, over-writing any previous values.",
    )
    lineMetrics: bool = Field(False, description="Whether to calculate line distance metrics. This is required for line layers that specify line-gradient values.")
    maxzoom: int = Field(18, description="Maximum zoom level at which to create vector tiles (higher means greater detail at high zoom levels).")
    promoteId: Optional[Union[str, Dict[str, str]]] = Field(
        description="A property to use as a feature id (for feature state). Either a property name, or an object of the form {<sourceLayer>: <propertyName>}."
    )
    tolerance: Optional[float] = Field(description="Douglas-Peucker simplification tolerance (higher means simpler geometries and faster performance).")


class Image(BaseModel):
    """
    An image source. The "url" value contains the image location.
    The "coordinates" array contains [longitude, latitude] pairs for the
    image corners listed in clockwise order: top left, top right, bottom right, bottom left.
    """

    type: Literal["image"] = "image"
    coordinates: Coords = Field(description="Corners of image specified in longitude, latitude pairs.")
    url: HttpUrl = Field(description="URL that points to an image.")


class Video(BaseModel):
    """
    An image source. The "url" value contains the image location.
    The "coordinates" array contains [longitude, latitude] pairs for the
    image corners listed in clockwise order: top left, top right, bottom right, bottom left.
    """

    type: Literal["video"] = "video"
    coordinates: Coords = Field(description="Corners of image specified in longitude, latitude pairs.")
    urls: List[HttpUrl] = Field(description="URL that points to an image.")


AnySource = Annotated[Union[Vector, Raster, RasterDem, GeoJson, Image, Video], Field(discriminator="type")]



import json
from importlib import resources

from maplibre.basemodel import Root

if __name__ == "__main__":

    style_text = resources.read_text("maplibre", "style.json")
    style = json.loads(style_text)

    root_two = Root.parse_obj(style)
    print(root_two)
    with resources.path("maplibre", "style_out.json") as p:
        with open(p, "w") as f:
            f.write(root_two.json(indent=4, exclude_unset=True))



from typing import Any, Dict
from unittest import TestCase

from maplibre import layer

# Create your tests here.
from . import sources


class LayersTestCase(TestCase):
    def test_layer(self):

        trunk_roads = layer.Layer(
            type="line",
            id="trunk_roads",
            source="roads",
            sourceLayer="trunk",
            layout={"line-cap": "round", "line-join": "round"},
            paint={"line-color": "#e9ac77", "line-width": {"base": 1.2, "stops": [[12, 1], [13, 3], [14, 4], [20, 15]]}},
        )

        trunk_roads.json()

    def test_reverse_layer(self):

        layer_detail = {
            "id": "landuse_residential",
            "type": "fill",
            "source": "openmaptiles",
            "source-layer": "landuse",
            "maxzoom": 8,
            "filter": ["==", "class", "residential"],
            "paint": {"fill-color": {"base": 1, "stops": [[9, "hsla(0, 3%, 85%, 0.84)"], [12, "hsla(35, 57%, 88%, 0.49)"]]}},
        }

        _layer = layer.Layer(**layer_detail)
        print(_layer)


class SourcesTestCase(TestCase):
    def test_vector_tiles(self):
        sources.Vector(**{"tiles": ["http://a.example.com/tiles/{z}/{x}/{y}.pbf", "http://b.example.com/tiles/{z}/{x}/{y}.pbf"], "maxzoom": 14}).json()

    def test_vector_tiles_json(self):
        sources.Vector(url="http://api.example.com/tilejson.json").json()

    def test_vector_mapbox(self):
        sources.Vector(url="mapbox://mapbox.mapbox-streets-v6").json()

    def test_vector_mapboxsatellite(self):
        sources.Vector(url="mapbox://mapbox.satellite", tileSize=256).json()

    def test_raster(self):
        sources.Raster(
            tiles=["http://a.example.com/wms?bbox={bbox-epsg-3857}&format=image/png&service=WMS&version=1.1.1&request=GetMap&srs=EPSG:3857&width=256&height=256&layers=example"],
            tileSize=256,
        ).json()

    def test_rasterdem(self):
        sources.RasterDem(url="mapbox://mapbox.terrain-rgb").json()

    def test_geojson(self):

        sources.GeoJson(
            data={"type": "Feature", "geometry": {"type": "Point", "coordinates": [-77.0323, 38.9131]}, "properties": {"title": "Mapbox DC", "marker-symbol": "monument"}}
        ).json()

    def test_image(self):
        sources.Image(
            url="https://maplibre.org/maplibre-gl-js-docs/assets/radar.gif", coordinates=[[-80.425, 46.437], [-71.516, 46.437], [-71.516, 37.936], [-80.425, 37.936]]
        ).json()

    def test_video(self):
        sources.Video(
            urls=["https://static-assets.mapbox.com/mapbox-gl-js/drone.mp4", "https://static-assets.mapbox.com/mapbox-gl-js/drone.webm"],
            coordinates=[
                [-122.51596391201019, 37.56238816766053],
                [-122.51467645168304, 37.56410183312965],
                [-122.51309394836426, 37.563391708549425],
                [-122.51423120498657, 37.56161849366671],
            ],
        ).json()


class DemoTilesTest(TestCase):
    """
    Reproduce the code at
    https://demotiles.maplibre.org/style.json
    in python files
    """

    style_text = ""
    style: Dict[str, Any] = {}

    @classmethod
    def setUpClass(cls) -> None:

        import json
        from importlib import resources

        cls.style_text = resources.read_text("maplibre", "style.json")
        cls.style = json.loads(cls.style_text)
        return super().setUpClass()

    def test_base(self):
        # Try the entire test file
        ...
        from .basemodel import Root

        root_two = Root.parse_obj(self.style)
        print(root_two)

    def test_demo_backgroundlayer(self):

        # Read the json
        # Try layers...
        layer.Layer(**self.style.get("layers")[0])


class OsmLibertyTest(TestCase):
    """
    Reproduce the code of the "osm Liberty" style
    """

    style_text = ""
    style: Dict[str, Any] = {}

    @classmethod
    def setUpClass(cls) -> None:

        import json
        from importlib import resources

        cls.style_text = resources.read_text("maplibre", "osm_liberty.json")
        cls.style = json.loads(cls.style_text)
        return super().setUpClass()

    def test_base(self):
        # Try the entire test file
        ...
        from .basemodel import Root

        root_two = Root.parse_obj(self.style)
        print(root_two)

    def test_demo_backgroundlayer(self):

        # Read the json
        # Try layers...
        layer.Layer(**self.style.get("layers")[0])



from dataclasses import dataclass
from typing import Dict, List, Optional, Protocol, Sequence, Tuple, Union

from django.db import models

P = Tuple[float, float]
L = Tuple[P]
ML = Tuple[L]

AnyGeom = Union[P, L, ML]


@dataclass
class GeoJsonGeometry:
    type: str  # = "name"
    coordinates: AnyGeom


@dataclass
class GeoJsonFeature:
    type: str  # "Feature"
    geometry: GeoJsonGeometry
    properties: Dict[str, str]


@dataclass
class GeoJsonFeatureCollection:
    type: str  # "FeatureCollection"
    features: List[GeoJsonFeature]


@dataclass
class GeoJsonSerializer:
    queryset: models.QuerySet
    geom_field: str = "geom"
    properties: Optional[Sequence[str]] = None

    def _to_feature(self, feature):
        geom = getattr(feature, self.geom_field)
        _t = geom.geom_type  # type: str
        _c = geom.coords  # type: AnyGeom
        return GeoJsonFeature(
            type="Feature",
            geometry=GeoJsonGeometry(type=_t, coordinates=_c),
            properties={k: getattr(feature, k) for k in self.properties or []},
        )

    def features(self):
        return [self._to_feature(feature) for feature in self.queryset]

    def to_collection(self):
        return GeoJsonFeatureCollection(type="FeatureCollection", features=self.features())


@dataclass
class MultiGeoJsonSerializer:
    serializers: Sequence[GeoJsonSerializer]

    def to_collection(self):

        features = []
        for s in self.serializers:
            features.extend(list(s.features()))

        return GeoJsonFeatureCollection(type="FeatureCollection", features=features)


class TileCache(Protocol):
    """
    Typing protocol for a generic "Cache" method
    The intention of this is to ensure that whichever cache method
    we use for an MvtTile, the cache has at least the methods which
    we access in our code
    """

    def get(self, key, default=None, version=None):
        """
        Fetch a given key from the cache. If the key does not exist, return
        default, which itself defaults to None.
        """
        pass

    def set(self, key, value, timeout=300, version=None):
        """
        Set a value in the cache. If timeout is given, use that timeout for the
        key; otherwise use the default cache timeout.
        """
        raise NotImplementedError("subclasses of BaseCache must provide a set() method")



from typing import Optional, Tuple

from pydantic import BaseModel, Field

"""
A style's sprite property supplies a URL template for loading small
images to use in rendering background-pattern,
fill-pattern, line-pattern,fill-extrusion-pattern
and icon-image style properties.
"""


class Sprite(BaseModel):
    width: int = 32
    height: int = 32
    x: int = 0
    y: int = 0
    pixelRatio: int = 1
    content: Optional[Tuple[float, float, float, float]] = Field(
        """An array of four numbers, with the first two specifying the left, top corner, and the last two specifying the right, bottom corner. If present, and if the icon uses icon-text-fit, the symbol's text will be fit inside the content box."""
    )
    stretchX: Optional[Tuple[float, float]] = Field(
        description="An array of two-element arrays, consisting of two numbers that represent the from position and the to position of areas that can be stretched."
    )
    stretchY: Optional[Tuple[float, float]]



# Generated by Django 3.2.8 on 2021-10-25 12:48

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djangostreetmap", "0005_osmislandsareas"),
    ]

    operations = [
        migrations.DeleteModel(
            name="PlanetOsmRoads",
        ),
        migrations.AlterField(
            model_name="osmislands",
            name="geom",
            field=django.contrib.gis.db.models.fields.MultiLineStringField(srid=3857),
        ),
    ]



from django.contrib.gis import admin

from djangostreetmap import models


@admin.register(models.SimplifiedLandPolygon)
class SimplifiedLandPolygonAdmin(admin.GeoModelAdmin):  # type: ignore
    pass



from pydantic import BaseModel

"""
A transition property controls timing for the interpolation between a
transitionable style property's previous value and new value.
A style's root transition property provides global
transition defaults for that style.
"""


class Transition(BaseModel):
    duration: int = 300
    delay: int = 0



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import AnyUrl, BaseModel, Field, validator

from .layer import Layer
from .light import Light
from .sources import AnySource


class Root(BaseModel):
    id: Optional[str]
    bearing: Optional[Union[int, float]] = Field(
        description="""Default bearing, in degrees. The bearing is the compass direction that is "up"; for example, a bearing of 90° orients the map so that east is up. This value will be used only if the map has not been positioned by other means (e.g. map options or user interaction)."""
    )
    center: Optional[Tuple[float, float]] = Field(
        description="""Default map center in longitude and latitude. The style center will be used only if the map has not been positioned by other means (e.g. map options or user interaction)."""
    )
    glyphs: Optional[AnyUrl] = Field(
        description="""A URL template for loading signed-distance-field glyph sets in PBF format. The URL must include {fontstack} and {range} tokens. This property is required if any layer uses the text-field layout property. The URL must be absolute, containing the scheme, authority and path components."""
    )
    light: Optional[Light] = Field(description="""The global light source.""")
    metadata: Optional[Dict[str, Any]] = Field(
        description="""Arbitrary properties useful to track with the stylesheet, but do not influence rendering. Properties should be prefixed to avoid collisions, like 'mapbox:'."""
    )
    name: Optional[str] = Field(description="""A human-readable name for the style.""")
    pitch: Optional[Union[int, float]] = Field(
        description="""Default pitch, in degrees. Zero is perpendicular to the surface, for a look straight down at the map, while a greater value like 60 looks ahead towards the horizon. The style pitch will be used only if the map has not been positioned by other means (e.g. map options or user interaction).""",
    )
    sources: Dict[str, AnySource] = Field(description="""Data source specifications.""")
    layers: List[Layer]
    sprite: Optional[AnyUrl] = Field(
        description="""A base URL for retrieving the sprite image and metadata. The extensions .png, .json and scale factor @2x.png will be automatically appended. This property is required if any layer uses the background-pattern, fill-pattern, line-pattern, fill-extrusion-pattern, or icon-image properties. The URL must be absolute, containing the scheme, authority and path components."""
    )
    transition: Optional[str] = Field(
        description="""A global transition definition to use as a default across properties, to be used for timing transitions between one value and the next when no property-specific transition is set. Collision-based symbol fading is controlled independently of the style's transition property."""
    )
    version: int = Field(8, description="""Style specification version number.""")

    zoom: Optional[float] = Field(
        description="""Default zoom level. The style zoom will be used only if the map has not been positioned by other means (e.g. map options or user interaction)."""
    )

    @validator("layers", allow_reuse=True)
    def layer_source_is_in_sources(cls, layers: List[Layer], values):
        """
        Ensure that the reference exists in the `sources` map
        """

        sources = values["sources"]  # type: Dict[str, AnySource]

        for layer in layers:

            if layer.type == "background":
                continue

            assert layer.source is not None

            source = sources.get(layer.source)

            assert source is not None, "A source-layer is required for vector types: source %s not found for layer %s. Sources are: %s" % (
                layer.source,
                layer.id,
                ", ".join(sources.keys()),
            )

            if source.type == "vector":
                assert layer.sourceLayer is not None, "A source-layer is required for vector types: no source is specified for layer %s" % (layer.id,)

            else:
                assert layer.sourceLayer is None, "A source-layer is not permitted except for vector types: remove sourceLayer for source %s" % (layer.id,)

        return layers



import hashlib
import logging
from dataclasses import asdict
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, Union

import sqlparse
from django.apps import apps
from django.contrib.gis.db.models.functions import Centroid, Transform
from django.db import connection
from django.db.models import Model
from django.http import HttpRequest, JsonResponse
from django.http.response import HttpResponse
from django.urls import reverse
from django.views import View
from django.views.generic.base import TemplateView
from osmflex.models import AmenityPoint, OsmLine, OsmPoint, OsmPolygon, RoadLine
from psycopg2 import sql

from djangostreetmap import models
from djangostreetmap.annotations import (
    GeoJsonSerializer,
    MultiGeoJsonSerializer,
    TileCache,
)
from djangostreetmap.functions import AsFeatureCollection, Intersects
from djangostreetmap.tilegenerator import MvtQuery, Tile
from maplibre import layer, sources
from maplibre.basemodel import Root
from maplibre.layer import Layer as L

from .timer import Timer

logger = logging.getLogger(__name__)


class ExampleMapView(TemplateView):
    template_name = "leaflet_tile_layers.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["map_view"] = "[-8.556416, 125.5524171], 14"
        return context


class TileLayerView(View):
    """
    Handle generation and views of tile layers
    For urls.py you likely want two entries for easier reversing in templates
    >>> path("mylayer/<int:zoom>/<int:x>/<int:y>.pbf", MyTileLayerView.as_view()), name="mylayer"),
    >>> path("mylayer", MyTileLayerView.as_view(), name="mylayer"),
    """

    layers: List[MvtQuery] = []
    tilecache: Optional[TileCache] = None

    def get_layers(self, tile: Tile) -> List[MvtQuery]:
        """
        Override for dynamically determining tile content
        based on tile properties (most likely zoom)
        """
        return self.layers

    def _generate_tile(self, tile: Tile) -> List[bytes]:

        params = asdict(tile)

        with connection.cursor() as cursor:
            tiles: List[bytes] = []
            for query_layer in self.get_layers(tile):
                query = query_layer.as_mvt()
                # Uncomment to see the SQL which is run
                # logger.info(query.as_string(cursor.cursor))
                with Timer(name="tile generator", logger=logger.info):
                    tile_response: Optional[Sequence] = None
                    key: Optional[str] = None

                    if self.tilecache:
                        key = hashlib.sha256(sqlparse.format(cursor.mogrify(query, params), reindent=True, keyword_case="upper").encode()).hexdigest()[:8]

                    if key:
                        content_bytes: Optional[bytes] = self.tilecache.get(key)
                        if content_bytes is not None:
                            logger.info("Cached tile returning")
                            tiles.append(content_bytes)
                            continue
                        logger.info("No tile cached")

                    try:
                        cursor.execute(query, params)  # type: ignore
                        tile_response = cursor.fetchone()
                    except Exception as E:
                        logger.error(f"{E}")
                        logger.info(query.as_string(cursor.cursor))
                    if not tile_response:
                        continue
                    content: memoryview = tile_response[0]
                    content_bytes: bytes = content.tobytes()
                    tiles.append(content_bytes)

                    if key:
                        self.tilecache.set(key, content_bytes)

            return tiles

    def get(self, request: HttpRequest, *args, **kwargs):
        with Timer(name="tile get", logger=logger.info):
            tile = Tile(**kwargs)  # Expect to receive zoom, x, and y in kwargs
            tiles = b"".join(self._generate_tile(tile))
            return HttpResponse(content=tiles, content_type="application/binary")


class BuildingPolygon(TileLayerView):
    def get_layers(self, tile: Tile):

        if tile.zoom < 14:
            return []

        return [
            MvtQuery(
                table="osmflex_buildingpolygon",
                attributes=["name", "osm_id", "osm_subtype"],
                calculated_attributes={"render_min_height": sql.SQL('COALESCE("height", "levels" * 4, 4)')},
                layer="buildings",
                pk="osm_id",
            )
        ]


class Roads(TileLayerView):
    def get_layers(self, tile: Tile):

        road_osm_types = [
            ("trunk", 2),
            ("steps", 12),
            ("road", 12),
            ("footway", 12),
            ("secondary", 7),
            ("tertiary", 9),
            ("secondary_link", 7),
            ("tertiary_link", 9),
            ("living_street", 12),
            ("pedestrian", 12),
            ("primary", 5),
            ("residential", 13),
            ("primary_link", 5),
            ("track", 12),
            ("motorway_link", 12),
            ("motorway", 5),
            ("service", 12),
            ("unclassified", 12),
            ("path", 12),
        ]  # type: List[Tuple[str, int]]

        # Determine which road types to include in the layer

        type_filter = sql.SQL("""{field} = ANY(ARRAY[{types}]::text[])""").format(
            field=sql.Identifier("osm_type"), types=sql.SQL(",").join([sql.Literal(rt) for rt, mz in road_osm_types if tile.zoom > mz])
        )

        roads = MvtQuery(
            table=RoadLine._meta.db_table,
            attributes=["name", "osm_type"],
            filters=[type_filter],
            layer="transportation",
            transform=False,
            pk="osm_id",
        )

        return [roads]


class Hospitals(View):
    def get(self, request: HttpRequest, *args, **kwargs):
        AmenityPoint = apps.get_model("osmflex", "AmenityPoint")
        AmenityPolygon = apps.get_model("osmflex", "AmenityPolygon")

        points = AmenityPoint.objects.filter(osm_type__in=["hospital", "clinic"]).annotate(geom_for_json=Transform("geom", 4326))
        centroids = AmenityPolygon.objects.filter(osm_type__in=["hospital", "clinic"]).annotate(geom_centroid=Transform(Centroid("geom"), 4326))

        response = MultiGeoJsonSerializer(
            [
                GeoJsonSerializer(points, "geom_for_json", ("name", "osm_type", "address")),
                GeoJsonSerializer(centroids, "geom_centroid", ("name", "osm_type", "address")),
            ]
        )

        return JsonResponse(data=asdict(response.to_collection()), content_type="application/json")


class Aeroways(View):
    def get(self, request: HttpRequest, *args, **kwargs):
        point = apps.get_model("osmflex", "InfrastructurePoint")
        polygon = apps.get_model("osmflex", "InfrastructurePolygon")

        points = point.objects.filter(osm_type__in=["aeroway"]).annotate(geom_for_json=Transform("geom", 4326))
        centroids = polygon.objects.filter(osm_type__in=["aeroway"]).annotate(geom_centroid=Transform(Centroid("geom"), 4326))

        response = MultiGeoJsonSerializer(
            [
                GeoJsonSerializer(points, "geom_for_json", ("name", "osm_type")),
                GeoJsonSerializer(centroids, "geom_centroid", ("name", "osm_type")),
            ]
        )

        return JsonResponse(data=asdict(response.to_collection()), content_type="application/json")


class LandLayer(TileLayerView):
    layers = [MvtQuery(table=models.SimplifiedLandPolygon._meta.db_table, layer="land")]


class PoiLayer(TileLayerView):
    layers = [
        MvtQuery(
            table=AmenityPoint._meta.db_table,
            attributes=["name"],
            # filters = sql.SQL("{} = {}").format(sql.Identifier("osm_type"), sql.Literal("school")),
            layer="school",
            pk="osm_id",
        )
    ]


class ModelFeatureCollectionView(View):
    """
    Container for OSM layer data
    """

    class Meta:
        proxy = True

    def geojson(
        self,
        model: Union[OsmPoint, OsmLine, OsmPolygon],
        filter_kwargs: Dict[str, Any],
        intersect: Type[Model],  # Note that we make some assumptions about "intersect" model here, see "Intersect" class
        **kwargs,
    ):
        """
        Returns a FeatureCollection of instances from "model" where the geometry
        intersects the area represented by this location profile

        For instance

        >>> OpenStreetMapData.objects.get(area__name="Highlands 1").geojson(AmenityPoint, osm_type="school")
        """
        # The OSM data import is in 3857. We transform to the geom of the
        # surrounding geometry, if applicable
        queryset = model.objects.all().annotate(geom_t=Transform("geom", 4326))
        queryset = queryset.annotate(in_area=Intersects(intersect)).filter(in_area=True)
        queryset = queryset.filter(**filter_kwargs)

        attributes = {field.name: field.name for field in model._meta.fields if field.name != "geom"}
        return queryset.aggregate(_=AsFeatureCollection("geom_t", **attributes))["_"]


class MapStyle(View):
    """
    Return a simple "mapbox"
    JSON view
    """

    def get(self, request: HttpRequest, *args, **kwargs):

        font = ["Roboto Regular"]
        # map = Root()
        road_source = sources.Vector(type="vector", tiles=[f"{request.scheme}://{request.get_host()}{reverse('djangostreetmap:roads')}{{z}}/{{x}}/{{y}}.pbf"])
        land_source = sources.Vector(type="vector", tiles=[f"{request.scheme}://{request.get_host()}{reverse('djangostreetmap:land')}{{z}}/{{x}}/{{y}}.pbf"])
        poi_source = sources.Vector(type="vector", tiles=[f"{request.scheme}://{request.get_host()}{reverse('djangostreetmap:poi')}{{z}}/{{x}}/{{y}}.pbf"])
        building_source = sources.Vector(type="vector", tiles=[f"{request.scheme}://{request.get_host()}{reverse('djangostreetmap:buildings')}{{z}}/{{x}}/{{y}}.pbf"])

        background = layer.Layer(id="background", type="background", paint={"background-color": "rgb(158,189,255)"})

        # Styles for every kind of road

        # Set the line, source and layout which are common properties of all roads
        road_common = {"type": "line", "source": "roads", "sourceLayer": "transportation", "layout": {"line-join": "round", "line-cap": "round"}}

        road_casing = [
            L(
                id="road_trunk_casing",
                filter=["all", ["==", "osm_type", "trunk"]],
                paint={"line-color": "#e9ac77", "line-width": {"base": 1.2, "stops": [[5, 0.4], [6, 0.7], [7, 1.75], [20, 22]]}},
                **road_common,
            ),
            L(
                id="road_primary_casing",
                filter=["all", ["==", "osm_type", "primary"]],
                paint={"line-color": "#e9ac77", "line-width": {"base": 1.2, "stops": [[5, 0.4], [6, 0.7], [7, 1.75], [20, 22]]}},
                **road_common,
            ),
            L(
                id="road_secondary_casing",
                filter=["all", ["==", "osm_type", "secondary"]],
                paint={"line-color": "#e9ac77", "line-width": {"base": 1.2, "stops": [[8, 1.5], [20, 17]]}},
                **road_common,
            ),
            L(
                id="road_tertiary_casing",
                filter=["all", ["==", "osm_type", "tertiary"]],
                paint={"line-color": "#e9ac77", "line-width": {"base": 1.2, "stops": [[8, 1.5], [20, 17]]}},
                **road_common,
            ),
            L(
                id="road_motorway_casing",
                filter=["all", ["==", "osm_type", "motorway"]],
                paint={"line-color": "#e9ac77", "line-width": {"base": 1.2, "stops": [[5, 0.4], [6, 0.7], [7, 1.75], [20, 22]]}},
                **road_common,
            ),
            L(
                id="road_residential_casing",
                filter=["all", ["==", "osm_type", "residential"]],
                paint={"line-color": "#cfcdca", "line-opacity": {"stops": [[12, 0], [12.5, 1]]}, "line-width": {"base": 1.2, "stops": [[12, 0.5], [13, 1], [14, 4], [20, 20]]}},
                **road_common,
            ),
            L(
                id="road_service_casing",
                filter=["all", ["==", "osm_type", "service"]],
                paint={"line-color": "#cfcdca", "line-width": {"base": 1.2, "stops": [[15, 1], [16, 4], [20, 11]]}},
                **road_common,
            ),
        ]

        # Road inner
        road_lines = [
            L(
                id="road_motorway",
                filter=["all", ["==", "osm_type", "motorway"]],
                paint={"line-color": {"base": 1, "stops": [[5, "hsl(26, 87%, 62%)"], [6, "#fc8"]]}, "line-width": {"base": 1.2, "stops": [[5, 0], [7, 1], [20, 18]]}},
                **road_common,
            ),
            L(
                id="road_trunk",
                filter=["all", ["==", "osm_type", "trunk"]],
                paint={"line-color": "#fea", "line-width": {"base": 1.2, "stops": [[5, 0], [7, 1], [20, 18]]}},
                **road_common,
            ),
            L(
                id="road_residential",
                filter=["all", ["==", "osm_type", "residential"]],
                paint={"line-color": "#fff", "line-width": {"base": 1.2, "stops": [[13.5, 0], [14, 2.5], [20, 18]]}},
                **road_common,
            ),
            L(
                id="road_service",
                filter=["all", ["==", "osm_type", "service"]],
                paint={"line-color": "#fff", "line-width": {"base": 1.2, "stops": [[13.5, 0], [14, 2.5], [20, 18]]}},
                **road_common,
            ),
            L(
                id="road_primary",
                filter=["all", ["==", "osm_type", "primary"]],
                paint={"line-color": "#fea", "line-width": {"base": 1.2, "stops": [[5, 0], [7, 1], [20, 18]]}},
                **road_common,
            ),
            L(
                id="road_secondary",
                filter=["all", ["==", "osm_type", "secondary"]],
                paint={"line-color": "#fea", "line-width": {"base": 1.2, "stops": [[6.5, 0], [8, 0.5], [20, 13]]}},
                **road_common,
            ),
            L(
                id="road_tertiary",
                filter=["all", ["==", "osm_type", "tertiary"]],
                paint={"line-color": "#fea", "line-width": {"base": 1.2, "stops": [[6.5, 0], [8, 0.5], [20, 13]]}},
                **road_common,
            ),
        ]

        # Road labels
        road_labels = [
            layer.Layer(
                id="road_highway_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "highway"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
            layer.Layer(
                id="road_trunk_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "trunk"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
            layer.Layer(
                id="road_primary_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "primary"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
            layer.Layer(
                id="road_secondary_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "secondary"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
            layer.Layer(
                id="road_tertiary_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "tertiary"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
            layer.Layer(
                id="road_residential_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "residential"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
            layer.Layer(
                id="road_service_label",
                type="symbol",
                source="roads",
                sourceLayer="transportation",
                filter=["all", ["==", "osm_type", "service"]],
                layout={
                    "symbol-placement": "line",
                    "text-anchor": "center",
                    "text-field": "{name}",
                    "text-font": font,
                    "text-offset": [0, 0.15],
                    "text-size": {"base": 1, "stops": [[13, 12], [14, 13]]},
                },
                paint={"text-color": "#765", "text-halo-blur": 0.5, "text-halo-width": 1},
            ),
        ]

        # POI (Points Of Interest)
        poi_layers = [
            layer.Layer(
                id="poi_school",
                type="symbol",
                source="poi",
                sourceLayer="school",
                layout={
                    "icon-image": "school_11",
                    "text-anchor": "top",
                    "text-field": "{name}",
                    "text-font": ["Roboto Condensed Italic"],
                    "text-max-width": 9,
                    "text-offset": [0, 0.6],
                    "text-size": 12,
                },
                paint={"text-color": "#666", "text-halo-blur": 0.5, "text-halo-color": "#ffffff", "text-halo-width": 1},
            )
        ]

        buildings = [
            layer.Layer(
                **{
                    "id": "building-3d",
                    "type": "fill-extrusion",
                    "source": "buildings",
                    "sourceLayer": "buildings",
                    "minzoom": 14,
                    "paint": {
                        "fill-extrusion-color": "hsl(35, 8%, 85%)",
                        "fill-extrusion-height": {"property": "render_min_height", "type": "identity"},
                        "fill-extrusion-opacity": 0.8,
                    },
                }
            )
        ]

        land_layer = layer.Layer(type="fill", id="land", source="land_polygons", sourceLayer="land", paint={"fill-color": "rgb(239,239,239)"})

        map = Root(
            sprite="https://maputnik.github.io/osm-liberty/sprites/osm-liberty",
            glyphs="https://api.maptiler.com/fonts/{fontstack}/{range}.pbf?key={key}",
            version=8,
            sources={
                "roads": road_source,
                "land_polygons": land_source,
                "poi": poi_source,
                "buildings": building_source,
            },
            layers=[background, land_layer, *road_casing, *road_lines, *road_labels, *poi_layers, *buildings],
        )

        return HttpResponse(map.json(exclude_unset=True, by_alias=True), content_type="application/json")



from http import HTTPStatus

from django.contrib.gis.geos import Point
from django.test import TestCase
from django.urls import reverse

from tests.models import BasicPoint


class MyTestCase(TestCase):
    def setUp(self) -> None:
        BasicPoint.objects.create(name="Pointy point point", geom=Point(1, 1))
        return super().setUp()

    def test_serialize(self):
        url = reverse("test_serializer")
        response = self.client.get(url)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_multi_serialize(self):
        url = reverse("test_multi_serializer")
        response = self.client.get(url)
        self.assertEqual(response.status_code, HTTPStatus.OK)



from dataclasses import dataclass, field
from typing import Dict, List, Sequence, Union

from django.contrib.gis.db.models import GeometryField
from django.db import connection
from psycopg2 import sql

# To time mvt queries uncomment the following
# from .timer import Timer


@dataclass
class Tile:
    """
    Simple dataclass representing a "tile" instance and additional metadata
    for tile generation: buffer, extent and margin
    http://postgis.net/docs/ST_TileEnvelope.html

    """

    zoom: int
    x: int
    y: int
    buffer: int = 64
    extent: int = 4096

    @staticmethod
    def tile_envelope() -> sql.Composable:
        return sql.SQL("ST_TileEnvelope(%(zoom)s, %(x)s, %(y)s)")

    @staticmethod
    def tile_envelope_margin() -> sql.Composable:
        """
        Returns the ST_TileEnvelope function with a margin
        Note that the margin is only available in postgis 3.1+
        """
        return sql.SQL("ST_TileEnvelope(%(zoom)s, %(x)s, %(y)s, margin => (%(buffer)s / %(extent)s))")


@dataclass
class MvtQuery:
    """
    This is a SQL query generator based on the example
    at https://postgis.net/docs/manual-dev/ST_AsMVT.html
    """

    table: str
    attributes: List[str] = field(default_factory=list)
    calculated_attributes: Dict[str, Union[sql.Composed, sql.SQL]] = field(default_factory=dict)
    filters: Sequence[sql.Composable] = field(default_factory=list)
    transform: bool = False  # Set to True if source srid is not 3857, but beware performance
    field: str = "geom"
    pk: str = "id"
    layer: str = "default"
    centroid: bool = False

    @property
    def json_attributes(self) -> sql.Composed:
        """
        Create a "JSONB_BUILD_OBJECT" clause
        """
        if not self.attributes and not self.calculated_attributes:
            return sql.SQL("").format()
        params = None
        for a in self.attributes:
            if not params:
                params = sql.Literal(a)
            else:
                params += sql.Literal(a)  # Name of the key is the same as the field
            params += sql.Identifier(a)  # The field to use as the value for the JSON

        for field, expression in self.calculated_attributes.items():
            if not params:
                params = sql.Literal(field)
            else:
                params += sql.Literal(field)  # Name of the key is the same as the field
            params += expression  # type: ignore # The expression to use as the value for the JSON

        composed_params = params.join(", ")  # type: ignore

        return sql.SQL(", jsonb_build_object({})").format(composed_params)

    @property
    def transformed_geom(self) -> sql.Composable:
        """
        Return an `ST_TRANSFORM` clause
        https://postgis.net/docs/ST_Transform.html
        """
        template = "{field}"
        if self.transform:
            template = f"ST_TRANSFORM({template}, 3857)"
        return sql.SQL(template).format(field=sql.Identifier(self.field))

    @property
    def centroid_wrap(self) -> sql.Composable:
        if self.centroid:
            return sql.SQL("ST_CENTROID({})").format(self.transformed_geom)
        return self.transformed_geom

    @property
    def alias(self) -> sql.Composable:
        """
        The name of the query alias to use for postgresql
        """
        return sql.Identifier(f"mvt_{self.layer}")

    @property
    def where(self) -> sql.Composable:
        return sql.SQL(" AND ") + sql.SQL(" AND ").join(self.filters) if self.filters else sql.SQL("")

    @property
    def as_mvtgeom(self) -> sql.Composable:

        return sql.SQL(
            """
            ST_AsMVTGeom(
                {cg},
                {e},
                extent => %(extent)s,
                buffer => %(buffer)s
            ) AS geom, {pk} {json_attributes}
            FROM {t}
            WHERE {where}
            """
        ).format(
            cg=self.centroid_wrap,
            e=Tile.tile_envelope(),
            t=sql.Identifier(self.table),
            # Properties of "self"
            pk=sql.Identifier(self.pk),
            json_attributes=self.json_attributes,
            where=self._feature_query,
        )

    @property
    def _feature_query(self) -> sql.Composable:
        """
        Return the sql required for the geometry query
        and other specified filters
        """
        return sql.SQL("""{g} && {m} {where}""").format(
            g=self.transformed_geom,
            m=Tile.tile_envelope(),
            # TODO: Re enable the margin when postgis >= 3.1
            # m=Tile.tile_envelope_margin(),
            where=self.where,
        )

    def as_mvt(self) -> sql.Composed:
        outer_query = sql.SQL("""WITH {alias} AS (SELECT {inner_query} ) SELECT ST_AsMVT( {alias}.*, {layer}, %(extent)s, 'geom', {pk}) FROM {alias}""")
        parameters = dict(alias=self.alias, layer=sql.Literal(self.layer), inner_query=self.as_mvtgeom, pk=sql.Literal(self.pk))
        return outer_query.format(**parameters)

    def debug(self) -> str:
        with connection.cursor() as cursor:
            return self.as_mvt().as_string(cursor.cursor)

    @classmethod
    def from_model(cls, model, *args, **kwargs) -> "MvtQuery":
        if "field" in kwargs:
            field = kwargs.pop("field")
        else:
            for field in model._meta.fields:
                if isinstance(field, GeometryField):
                    field = field.db_column or field.attname
                    break
        assert field, f"No geometry field could be identified for {model}"

        if "attributes" in kwargs:
            attributes = kwargs.pop("attributes")
        else:
            attributes = [f.db_column or f.attname for f in model._meta.fields if not isinstance(f, GeometryField)]

        if "pk" in kwargs:
            pk = kwargs["pk"]
        else:
            for pk_field_candidate in model._meta.fields:
                if pk_field_candidate.primary_key:
                    pk = pk_field_candidate.db_column or pk_field_candidate.attname
                    break
        assert pk, f"No primary key field could be identified for {model}"

        return cls(
            table=model._meta.db_table,
            attributes=attributes,
            field=field,
            transform=model._meta.get_field(field).srid != 3857,
            pk=pk,
            layer=kwargs.pop("layer", model._meta.model_name),
            **kwargs,
        )
